from typing import Tuple, List

import matplotlib.pyplot as plt
import tensorflow as tf

from histomics_detect.boxes.transforms import filter_edge_boxes
from histomics_detect.anchors.create import create_anchors
from histomics_detect.metrics import greedy_iou_mapping
from histomics_detect.metrics.iou import iou
from histomics_detect.models.model_utils import extract_data
from histomics_detect.models.lnms_model import LearningNMS
from histomics_detect.models.compression_network import CompressionNetwork


def plot_inference(rgb: tf.Tensor, boxes: tf.Tensor, nms_output: tf.Tensor, rpn_boxes: tf.Tensor,
                   save_fig: bool = False, fig_path: str = 'plot_of_inference.png', filter_edge: bool = False,
                   filter_margin: int = 32, threshold: float = 0.5, figsize: Tuple[int, int] = (20, 20),
                   gt_colors: Tuple[str, str] = ('g', 'r'), pred_colors: Tuple[str, str] = ('g', 'r'),
                   pred_size: int = 300, print_prediction_numbers: bool = True, show_axis: bool = True,
                   ax=plt, is_multi: bool = False) -> None:
    """
    calculates the statistic of previously run output and plots it

    Parameters
    ----------
    rgb: array
        Image for display with imshow.
    boxes: tensor (float32)
        ground truth boxes
        shape: G x 4
    nms_output: tensor (float32)
        objectiveness scores corresponding to the predicted boxes after lnms processing
        shape: N x 1
    rpn_boxes: tensor
        N x 4 tensor where each row contains the x,y location of the upper left
        corner of each regressed box and its width and height in that order.
    save_fig: bool
        true if picture should be saved
    fig_path: str
        path to save figure at
    filter_edge: bool
        true -> filters boxes and predictions too close to the border of image
    filter_margin: int
        min distance to border not to be filtered out
    threshold: float
    figsize: Tuple[int, int]
        dimensions of figure
    gt_colors: Tuple[str, str]
        color of predicted ground truth, color of not predicted ground truth (false negative)
    pred_colors: Tuple[str, str]
        color of true positive prediction, color of false positive prediction
    pred_size: int
        size of dot symbolizing a prediction
    print_prediction_numbers: bool
        true -> print statistic of tp, fp, fn
    show_axis: bool

    Returns
    -------
    None

    """
    scores = tf.reshape(nms_output, -1)
    condition = tf.greater(scores, threshold)
    rpn_boxes = tf.squeeze(tf.gather(rpn_boxes, tf.where(condition)))

    if filter_edge:
        boxes, _ = filter_edge_boxes(boxes, tf.shape(rgb)[1], tf.shape(rgb)[0], filter_margin)

    try:
        ious = iou(rpn_boxes, boxes)

        tp, fp, fn, tp_list, fp_list, fn_list = greedy_iou_mapping(ious, 0.18)

        positive_pred = tf.reshape(tf.gather(rpn_boxes, tp_list[:, 0]), (-1, 4))
        negative_pred = tf.reshape(tf.gather(rpn_boxes, fp_list), (-1, 4))  # tf.where(labels == 0)
        negative_boxes = tf.reshape(tf.gather(boxes, fn_list), (-1, 4))

        if filter_edge:
            negative_boxes, _ = filter_edge_boxes(negative_boxes, tf.shape(rgb)[1], tf.shape(rgb)[0], 
                                                  filter_margin, tf.constant(False, tf.bool))
            positive_pred, _ = filter_edge_boxes(positive_pred, tf.shape(rgb)[1], tf.shape(rgb)[0], 
                                                 filter_margin, tf.constant(False, tf.bool))
            negative_pred, _ = filter_edge_boxes(negative_pred, tf.shape(rgb)[1], tf.shape(rgb)[0], 
                                                 filter_margin, tf.constant(False, tf.bool))

    except:
        positive_pred = tf.constant([])
        negative_pred = tf.constant([])
        negative_boxes = boxes

    if print_prediction_numbers:
        print(f'tp: {tf.shape(positive_pred)[0]}, fp: {tf.shape(negative_pred)[0]}, fn: '
              f'{tf.shape(boxes)[0] - tf.shape(positive_pred)[0]}')

    if not is_multi:
        fig = plt.figure(figsize=figsize)
    ax.imshow(tf.cast(rgb, tf.int32))

    _plot_boxes_multi_plot(boxes, gt_colors[0], ax)  # '#b9ff38'
    _plot_boxes_multi_plot(negative_boxes, gt_colors[1], ax)  # '#b9ff38'

    if tf.size(negative_pred) > 0:
        negative_pred_d = negative_pred[:, :2] + negative_pred[:, 2:] / 2
        ax.scatter(negative_pred_d[:, 0], negative_pred_d[:, 1], color=pred_colors[1], s=pred_size)
    if tf.size(positive_pred) > 0:
        positive_pred_d = positive_pred[:, :2] + positive_pred[:, 2:] / 2
        ax.scatter(positive_pred_d[:, 0], positive_pred_d[:, 1], color=pred_colors[0], s=pred_size)

    if not show_axis:
        ax.axis('off')

    if not is_multi:
        plt.show()
    if save_fig and not is_multi:
        fig.savefig(fig_path)


def _run_model(data, model):
    """
    runs the lnms model

    Parameters
    ----------
    data:
        model input
    model:
        network

    Returns
    -------
    rgb: array
        Image for display with imshow.
    boxes: tensor (float32)
        ground truth boxes
        shape: G x 4
    rpn_boxes: tensor
        N x 4 tensor where each row contains the x,y location of the upper left
        corner of each regressed box and its width and height in that order.
    nms_output2: tensor (float32)
        objectiveness scores corresponding to the predicted boxes after lnms processing
        shape: N x 1
    """
    rgb, boxes, _ = data
    print(_)

    width, height, _ = tf.shape(rgb)
    model.anchors = create_anchors(model.anchor_px, model.field, height, width)

    norm, boxes, sample_weight = extract_data(data)
    features, rpn_boxes, scores = model.extract_boxes_n_scores(norm)

    comp_features = model.compression_net(features)
    interpolated = model._interpolate_features(comp_features, rpn_boxes)

    interpolated = tf.concat([scores, tf.reshape(interpolated, [tf.shape(interpolated)[0], -1])], axis=1)
    nms_output2 = model.net((interpolated, rpn_boxes), training=True)
    nms_output2 = tf.expand_dims(nms_output2[:, 0], axis=1)

    return rgb, boxes, rpn_boxes, nms_output2


def _plot_boxes_multi_plot(boxes, color, ax):
    """
    plots boxes on specific ax

    Parameters
    ----------
    boxes:
        boxes to plot
    color:
        box color
    ax:
        axis to plot boxes on or plt

    Returns
    -------

    """
    x, y, w, h = tf.split(boxes, 4, axis=1)
    for i, (xi, yi, wi, hi) in enumerate(zip(x, y, w, h)):
        ax.plot([xi, xi + wi, xi + wi, xi, xi],
                [yi, yi, yi + hi, yi + hi, yi],
                color=color)


def run_plot(validation_data, model: tf.keras.Model, index: int = 0, save_fig: bool = False,
             fig_path: str = 'plot_of_inference.png', filter_edge: bool = False, filter_margin: int = 32,
             threshold: float = 0.5, figsize: Tuple[int, int] = (20, 20), gt_colors: Tuple[str, str] = ('g', 'r'),
             pred_colors: Tuple[str, str] = ('g', 'r'), pred_size: int = 300, print_prediction_numbers: bool = True,
             show_axis: bool = True) -> None:
    """
    runs model ond plots the inference

    Parameters
    ----------
    save_fig: bool
        true if picture should be saved
    fig_path: str
        path to save figure at
    filter_edge: bool
        true -> filters boxes and predictions too close to the border of image
    filter_margin: int
        min distance to border not to be filtered out
    threshold: float
    print_prediction_numbers: bool
        true -> print statistic of tp, fp, fn
    figsize: Tuple[int, int]
        dimensions of figure
    gt_colors: Tuple[str, str]
        color of predicted ground truth, color of not predicted ground truth (false negative)
    pred_colors: Tuple[str, str]
        color of true positive prediction, color of false positive prediction
    pred_size: int
        size of dot symbolizing a prediction
    show_axis: bool
    validation_data
    model: tf.keras.Model
        lnms model to run
    index: int
        index of validation image that should be run and plotted

    Returns
    -------

    """

    counter = 0
    for data in validation_data:
        if counter < index:
            counter += 1
            continue

        rgb, boxes, rpn_boxes, nms_output2 = _run_model(data, model)

        plot_inference(rgb, boxes, nms_output2, rpn_boxes, save_fig, fig_path, filter_edge, filter_margin, threshold,
                       figsize, gt_colors, pred_colors, pred_size, print_prediction_numbers, show_axis)
        break


def plot_multiple_outputs(configs: dict, validation_data: tf.data.Dataset, model_paths: List[str], variable: str,
                          model_variations: List, index: int, faster_model: tf.keras.Model, titles: List[str],
                          fig_path: str = 'figures/plot_of_inferences.png') -> None:
    """
    plots multiple outputs of lnms where one hyperparamter changes

    Parameters
    ----------
    configs: dict
        configurations as dictionary
    validation_data:
        validation dataset
    model_paths:
        list of paths to model weights
    variable:
        configs variable that changes
    model_variations:
        list of the values of the variable that changes
    index:
    faster_model:
        faster r-cnn model with loaded weights
    titles:
        list of titles for each plot
    fig_path:
        save fig here

    Returns
    -------

    """
    counter = 0
    for data in validation_data:
        if counter == index:
            data = data
            break
        counter += 1
    #     validation_data = list(validation_data.as_numpy_iterator())
    #     data = validation_data[index]

    rgb, boxes, _ = data

    fig, axs = plt.subplots(1, len(model_paths) + 1, sharey=True, figsize=(30, 20))

    regressions = faster_model(rgb, tau=0.5, nms_iou=0.3)

    scores = tf.ones((tf.shape(regressions)[0], 1))
    plot_inference(rgb, boxes.to_tensor(), scores, regressions, filter_edge=True, is_multi=True, ax=axs[0],
                   pred_size=50, show_axis=False)
    axs[0].title.set_text("NMS")

    compression_net = CompressionNetwork(configs['feature_size'], configs['anchor_size'], faster_model.backbone)

    for i, (path, model_variation) in enumerate(zip(model_paths, model_variations)):
        ax = axs[i + 1]
        ax.title.set_text(titles[i])

        copied_config = configs.copy()
        copied_config[variable] = model_variation

        model = LearningNMS(configs, faster_model.rpnetwork, faster_model.backbone, compression_net.compression_layers,
                            [configs['width'], configs['height']], )
        model.compile(optimizer=tf.keras.optimizers.Adam(learning_rate=1e-4))

        model.load_weights(path)

        rgb, boxes, rpn_boxes, nms_output2 = _run_model(data, model)

        plot_inference(rgb, boxes, nms_output2, rpn_boxes, False, '', filter_edge=True, ax=ax, is_multi=True,
                       pred_size=50, show_axis=False)

    plt.subplots_adjust(wspace=0.01, hspace=0)
    plt.show()
    fig.savefig(fig_path)



from histomics_detect.networks.transfer_layers import transfer_layers
import tensorflow as tf


def pretrained(name):
    """
    Loads backbone keras model from tf.keras.applications.
    
    Parameters
    ----------
    name : string
        Name of pre-trained network to load. Case insensitive.

    Returns
    -------
    model : tf.keras.Model
        A pre-trained tf.keras.Model network.
    preprocessor : function
        Function used for transforming model inputs.
    """
    
    if str.lower(name) == 'resnet50':
        model = tf.keras.applications.resnet.ResNet50(
            include_top=False, weights='imagenet', input_tensor=None,
            input_shape=(None, None, 3), pooling=None)
        preprocessor = tf.keras.applications.resnet.preprocess_input
    elif str.lower(name) == 'resnet101':
        model = tf.keras.applications.resnet.ResNet101(
            include_top=False, weights='imagenet', input_tensor=None,
            input_shape=(None, None, 3), pooling=None)
        preprocessor = tf.keras.applications.resnet.preprocess_input
    elif str.lower(name) == 'resnet152':
        model = tf.keras.applications.resnet.ResNet152(
            include_top=False, weights='imagenet', input_tensor=None,
            input_shape=(None, None, 3), pooling=None)
        preprocessor = tf.keras.applications.resnet.preprocess_input
    elif str.lower(name) == 'resnet50v2':
        model = tf.keras.applications.resnet_v2.ResNet50V2(
            include_top=False, weights='imagenet', input_tensor=None,
            input_shape=(None, None, 3), pooling=None)
        preprocessor = tf.keras.applications.resnet_v2.preprocess_input
    elif str.lower(name) == 'resnet101v2':
        model = tf.keras.applications.resnet_v2.ResNet101V2(
            include_top=False, weights='imagenet', input_tensor=None,
            input_shape=(None, None, 3), pooling=None)
        preprocessor = tf.keras.applications.resnet_v2.preprocess_input
    elif str.lower(name) == 'resnet152v2':
        model = tf.keras.applications.resnet_v2.ResNet152V2(
            include_top=False, weights='imagenet', input_tensor=None,
            input_shape=(None, None, 3), pooling=None)
        preprocessor = tf.keras.applications.resnet_v2.preprocess_input
    else:
        raise ValueError("Network name not recognized")
    
    return model, preprocessor


def residual(model, preprocessor, blocks, stride=None):
    """
    Creates a feature extraction backbone from a tf.keras.applications resnet model.
    Allows user to select the number of residual blocks to keep and to set the convolution
    stride. Optionally merges the model with a preprocessor function.
        
    Parameters
    ----------
    model : tf.keras.Model
        A resnet keras Model obtained from tf.keras.applications. Can be loaded using 'pretrained()'.
    preprocesor : function
        Function used to transform input images. Can be loaded using 'pretrained()'.
    blocks : int
        The desired number of residual blocks to keep from the input model. Terminal blocks
        > blocks are truncated from the output model.
    stride : int
        The desired stride for the first convolution. Default value None does not alter stride.
    
    Returns
    -------
    backbone : tf.keras.Model
        A keras model with desired number of blocks, stride, and preprocessing capability.
    """

    #sweep layers to identify activation layer of the block 'blocks'
    add = False
    block = 0
    for (i,layer) in enumerate(model.layers):

        #change state if add layer encountered and continue with next layer
        if type(layer) == tf.keras.layers.Add:
            add = True
            continue

        #look for activation or batch norm layer following add layer
        if add:
            if (type(layer) == tf.keras.layers.Activation): #resnet v1 - keep activation
                block = block + 1
                add = False
                if block == blocks:
                    terminal = i
                    break
            elif (type(layer) == tf.keras.layers.BatchNormalization): #resnet v2 - keep add
                block = block + 1
                add = False
                if block == blocks:
                    terminal = i-1
                    break
            else:
                add = False

    #replace input layer
    input = tf.keras.Input(shape=(None, None, 3))

    #duplicate zero padding layer (layer 1)
    padding = tf.keras.layers.serialize(model.layers[1])
    padding = tf.keras.layers.deserialize(padding)

    #modify stride of first convolutional layer (layer 2)
    weights = model.layers[2].get_weights()
    name = model.layers[2].name
    if stride is None:
        stride = model.layers[2].strides
    conv = tf.keras.layers.serialize(model.layers[2])
    conv['config']['strides'] = (stride, stride)
    conv = tf.keras.layers.deserialize(conv)
    
    #apply preprocessor
    normalized = preprocessor(input)

    #compose model
    features = transfer_layers(model.layers[3:terminal+1],
                                   'b', conv(padding(normalized)))   
    backbone = tf.keras.Model(inputs=input, outputs=features)
    
    #set weights for first convolution
    for (i, layer) in enumerate(backbone.layers):
        if layer.name == name:
            backbone.layers[i].set_weights(weights)

    return backbone



"""
This package contains Keras Model subclasses for detection
"""

# make functions available at the package level using shadow imports
from .faster_rcnn import FasterRCNN, faster_rcnn_config
from .lnms_loss import normal_loss, xor_loss, clustering_loss, paper_loss
from .compression_network import CompressionNetwork
from .lnms_model import LearningNMS
from .block_model import BlockModel
from .model_utils import extract_data
# list out things that are available for public use
__all__ = (

    # functions and classes of this package
    'FasterRCNN',
    'faster_rcnn_config',
    'normal_loss',
    'xor_loss',
    'clustering_loss',
    'paper_loss',
    'CompressionNetwork',
    'LearningNMS',
    'BlockModel',
    'extract_data'
)



import tensorflow as tf


@tf.function
def _box_crop(corner, length, window):
    """Applies a crop to a sequence of boxes to remove portions falling outside
    a defined region.
    
    This function is applied to each dimension independently. Given the region
    with edge length 'window', this crops the boxes to clip portions that lie 
    outside the top/left or bottom right of the region. Boxes with zero overlap
    will have their height/width set to zero.
        
    Parameters
    ----------
    corner: tensor (float32)
        Coordinates of left or top of boxes.
    length: tensor (float32)
        Width or height of boxes.
    window: float32
        Width or height of region, assume upper/left corner is at 0.
        
    Returns
    -------
    corner_crop: tensor (float32)
        1D tensor containing coordinate of upper/left of cropped boxes.
    intersection: tensor (float32)
        1D tensor containing height/width of cropped boxes.
    """

    corner_crop = tf.minimum(tf.maximum(corner, 0), window)
    length_crop = tf.minimum(tf.maximum(corner+length, 0), window) - corner_crop

    return corner_crop, length_crop


@tf.function
def flip(rgb, boxes):
    """Randomly flips an image and ground truth boxes along horizontal and/or 
    verical axis.
    
    This function is used for data augmentation that consistently flips the 
    image array as well as transposing the box coordinates and width/height at
    random.
        
    Parameters
    ----------
    corner: tensor (float32)
        Coordinates of left or top of boxes.
    length: tensor (float32)
        Width or height of boxes.
    window: float32
        Width or height of region, assume upper/left corner is at 0.
        
    Returns
    -------
    corner_crop: tensor (float32)
        1D tensor containing coordinate of upper/left of cropped boxes.
    intersection: tensor (float32)
        1D tensor containing height/width of cropped boxes.
    """

    #condense ragged tensor
    [x, y, w, h] = tf.unstack(boxes.to_tensor(), num=4, axis=1)

    if tf.random.uniform([1])[0] > 0.5:
    
        #flip image
        rgb = tf.image.flip_up_down(rgb)

        #flip coordinates
        y = tf.cast(tf.shape(rgb)[0], tf.float32) - y - h

    if tf.random.uniform([1])[0] > 0.5:
    
        #flip image
        rgb = tf.image.flip_left_right(rgb)

        #flip coordinates
        x = tf.cast(tf.shape(rgb)[1], tf.float32) - x - w

    #re-stack
    boxes = tf.RaggedTensor.from_tensor(tf.stack([x, y, w, h], axis=1))

    return rgb, boxes


@tf.function
def crop(rgb, boxes, width, height, min_fraction=0.5):
    """Randomly crops a portion of the input image and ground truth boxes.
    
    This function is used for data augmentation and randomly crops a
    height x width portion of the input and corresponding ground truth
    boxes. Removes cropped boxes that do not have at least 'min_fraction'
    of their area contained within the cropped region to clean up
    boundaries. To avoid producing empty output that does not contain any
    ground truth boxes, we pick a single box at random and ensure that this 
    is wholly contained within the crop.
        
    Parameters
    ----------
    rgb: tensor
        The 2D or 3D input image tensor.
    boxes: tensor (float32)
        M x 4 tensor where each row contains the x,y location of the upper left
        corner of a ground truth box and its width and height in that order.
    width: int32
        Width of the cropped output image.
    height: int32
        Height of the cropped output image.
    min_fraction: float32
        Threshold for percent of original area used to retain cropped boxes.
        
    Returns
    -------
    crop: tensor
        Cropped height x width image.
    boxes: tensor (float32)
        N x 4 tensor containing cropped boxes where each row contains the x,y 
        location of the upper left corner of a ground truth box and its width 
        and height in that order. Each box is at least min_fraction percent
        of its original uncropped size.
    """

    #modify shapes
    width = tf.minimum(width, tf.shape(rgb)[1])
    height = tf.minimum(height, tf.shape(rgb)[0])

    #condense ragged tensor
    [x, y, w, h] = tf.unstack(boxes.to_tensor(), num=4, axis=1)

    #identify candidate cells to include in cropped roi
    c_width = tf.less(w, tf.cast(width, tf.float32))
    c_height = tf.less(h, tf.cast(height, tf.float32))
    c_left = tf.greater_equal(x, 0.0)
    c_top = tf.greater_equal(y, 0.0)
    c_right = tf.less(x+w, tf.cast(tf.shape(rgb)[1], tf.float32))
    c_bottom = tf.less(y+h, tf.cast(tf.shape(rgb)[0], tf.float32))
    conditions = tf.stack([c_width, c_height, c_left,
                           c_top, c_right, c_bottom], axis=1)
    include = tf.cast(tf.where(tf.reduce_all(conditions, axis=1)), tf.int32)
    x = tf.gather(x, include)
    y = tf.gather(y, include)
    w = tf.gather(w, include)
    h = tf.gather(h, include)

    #select box/cell at random
    selected = tf.random.uniform([1], 0, tf.size(x)-1, dtype=tf.int32)[0]
  
    #Determine range of crop limits for upper left corner of random 
    #(height, width) fields that include box. Calculate this in floats to deal
    #with edge cases like floor(x[selected])=0. 
    lower_x = tf.maximum(x[selected] + w[selected] - tf.cast(width, tf.float32), 0)
    upper_x = tf.minimum(x[selected], tf.cast(tf.shape(rgb)[1] - width,
                                              tf.float32))
    lower_y = tf.maximum(y[selected] + h[selected] - tf.cast(height, tf.float32), 0)
    upper_y = tf.minimum(y[selected], tf.cast(tf.shape(rgb)[0] - height,
                                              tf.float32))

    #sample upper left corner of roi
    x_crop = tf.random.uniform([1], lower_x[0], upper_x[0], tf.float32)[0]
    y_crop = tf.random.uniform([1], lower_y[0], upper_y[0], tf.float32)[0]
    x_crop = tf.cast(tf.round(x_crop), tf.int32)
    y_crop = tf.cast(tf.round(y_crop), tf.int32) 

    #crop 
    crop = tf.image.crop_to_bounding_box(rgb, y_crop, x_crop, 
                                         tf.cast(height, tf.int32),
                                         tf.cast(width, tf.int32))

    #translate boxes and apply crop
    x = x - tf.cast(x_crop, tf.float32)
    y = y - tf.cast(y_crop, tf.float32)
    x, wc = _box_crop(x, w, tf.cast(width, tf.float32))
    y, hc = _box_crop(y, h, tf.cast(height, tf.float32))

    #calculate cropped box area, proportion of box in cropped region
    proportion = tf.divide(tf.multiply(wc, hc), tf.multiply(w, h))

    #assign outputs
    mask = tf.greater_equal(proportion, min_fraction)
    x = tf.boolean_mask(x, mask)
    y = tf.boolean_mask(y, mask)
    w = tf.boolean_mask(wc, mask)
    h = tf.boolean_mask(hc, mask)

    boxes = tf.RaggedTensor.from_tensor(tf.stack((x,y,w,h), axis=1))

    return crop, boxes
  

@tf.function
def jitter(boxes, percent=0.05):
    """Randomly displaces bounding boxes using uniform noise proportional
    to a percentage of box dimensions.
    
    This function is used for data augmentation and adds noise proportional
    to bounding box dimensions to the box coordinates.
        
    Parameters
    ----------
    boxes: tensor (float32)
        M x 4 tensor where each row contains the x,y location of the upper left
        corner of a ground truth box and its width and height in that order.
    percent: float
        Range of noise applied to bounding boxes. Default value of 0.05
        corresponds to up to a +/- 5 percent displacement.
        
    Returns
    -------
    boxes: tensor (float32)
        M x 4 tensor where the x,y locations of each box have been displaced by 
        up to 'percent' of the box width and height respectively.
    """
    
    #condense ragged tensor
    [x, y, w, h] = tf.unstack(boxes.to_tensor(), num=4, axis=1)
    
    #generate noise vectors
    x_noise = tf.random.uniform(tf.shape(x), -percent, percent, tf.float32)
    y_noise = tf.random.uniform(tf.shape(y), -percent, percent, tf.float32)
    
    #modify locations
    x = x + tf.multiply(x_noise, w)
    y = y + tf.multiply(y_noise, h)
    
    #stack result and return
    boxes = tf.RaggedTensor.from_tensor(tf.stack((x,y,w,h), axis=1))
    
    return boxes


@tf.function
def shrink(boxes, percent=0.05):
    """Randomly resizes bounding boxes proportional to box dimensions.
    
    This function is used for data augmentation.
        
    Parameters
    ----------
    boxes: tensor (float32)
        M x 4 tensor where each row contains the x,y location of the upper left
        corner of a ground truth box and its width and height in that order.
    percent: float
        Range of percentages used to resize bounding boxes. Default value of 0.05
        corresponds to up to a +/- 5 percent change in size.
        
    Returns
    -------
    boxes: tensor (float32)
        M x 4 tensor where the width and height of each box have been changed by 
        up to 'percent' of the box width and height respectively.
    """
    
    #condense ragged tensor
    [x, y, w, h] = tf.unstack(boxes.to_tensor(), num=4, axis=1)
    
    #generate noise vectors
    w_noise = tf.random.uniform(tf.shape(x), 1.0-percent, 1.0+percent, tf.float32)
    h_noise = tf.random.uniform(tf.shape(y), 1.0-percent, 1.0+percent, tf.float32)
    
    #modify dimensions
    w = tf.multiply(w_noise, w)
    h = tf.multiply(h_noise, h)
    
    #stack result and return
    boxes = tf.RaggedTensor.from_tensor(tf.stack((x,y,w,h), axis=1))
    
    return boxes



import tensorflow as tf


def transfer_layers(layers, prefix, input):
    #This function copies layers from the input models into a single
    #model with appropriate name scopes to help separate variables by
    #shared, predictor, and discriminator sub-networks. Supports feed 
    #forward networks with skip layers and branching architectures. 
    #layers - a list of tf.keras.layers
    #prefix - string used for scoping variable and layer names
    #input - optional model input from other subnetwork, if 'None'
    #        model.layers[0] must be tf.keras.layers.Input.

    #check input layer type and discard input layer if present
    if layers[0].__class__.__name__ == 'InputLayer':
        layers = layers[1:]

    #history of layer output names and their indices in model.layers
    history = {}

    #list of transferred model layer output, used to feed later layers
    outputs = []

    #iterate through layers, transferring configurations and weights
    for i, layer in enumerate(layers):
        
        #layer output name is key, index in 'output' is value
        history[layer.output.name] = i

        #generate new layer w/ configuration w/o weights
        serialized = tf.keras.layers.serialize(layer)
        serialized['config']['name'] = prefix + '/' + serialized['config']['name']
        transfered = tf.keras.layers.deserialize(serialized)

        #set new layer inputs and outputs
        if i == 0:
            outputs.append(transfered(input))
        else:
            if type(layer.input) is list:
                inputs = [outputs[history[input.name]] for input in layer.input]
                outputs.append(transfered(inputs))
            else:
                inputs = outputs[history[layer.input.name]]
                outputs.append(transfered(inputs))

        #copy weights to new layer
        transfered.set_weights(layer.get_weights())

    return outputs[-1]



import tensorflow as tf


def rpn(features, n_anchors, kernels=[3], dimensions=[256],
        activations=['relu']):
    """
    Produces a region proposal network that consumes 3D feature maps from the
    backbone, and produces objectness scores and box regressions used for
    region proposal.
    
    Parameters
    ----------
    features : int
        The number of features in the 3D feature map generated by the backbone.
    n_anchors : int
        The number of anchors at each grid point.
    kernels : array_like (int)
        A list or array defining the number of convolutional layers and kernel sizes
        of each layer. A value of [3, 2] will produce a two-layer cascade, with the
        first layer having 3x3 kernels and the second layer having 2x2 kernels. These 
        kernels convolve over the 3D feature map where the grid spacing corresponds to
        the backbone field size. Default value [3].
    dimensions : array_like (int)
        A list or array defining the number of kernels in each convolutional layer. 
        Must be same length as kernels. Default value [256].
    activations : array_like (string)
        A list of strings defining the activations to use in each convolution layer.
        Must be same length as kernels. Default value ['relu'].
    
    Returns
    -------
    rpn : tf.keras.Model
        The region proposal network. Consumes a 3D tensor and produces objectness
        scores and box regressions.
    """
    
    #create input layer
    rpn_input = tf.keras.Input(shape=(None, None, features))

    #build region-proposal convolutional layers
    for i, (kernel, dimension, activation) in \
        enumerate(zip(kernels, dimensions, activations)):
        if i == 0:
            layer_input = rpn_input
        else:
            layer_input = x  
        x = tf.keras.layers.Conv2D(dimension, kernel, padding='same',
                                   activation=activation)(layer_input)

    #build regression and objectness outputs using 1D convolutions
    regression = tf.keras.layers.Conv2D(4*n_anchors, 1, padding='same',
                                        activation='linear')(x)
    objectness = tf.keras.layers.Conv2D(2*n_anchors, 1, padding='same',
                                        activation='linear')(x)

    #create rpn model
    rpn = tf.keras.Model(inputs=rpn_input, 
                         outputs=[objectness, regression])

    return rpn



from typing import Union
import tensorflow as tf
import tensorflow.keras.backend as kb

from histomics_detect.metrics.iou import iou


def assemble_single_neighborhood(anchor_id: int, interpolated: tf.Tensor, neighborhood_indeces: tf.Tensor,
                                 neighborhood_additional_info: tf.Tensor, use_image_features: bool = True,
                                 use_joint_features: bool = False, use_cross_feature: bool = False) \
        -> tf.float32:
    """
    DEPRECIATED: unused

    assembles the prediction representations for a neighborhood of a single prediction

    the neighborhood assembly of a single prediction consists of
    - collecting the neighboring prediction representation (initially the interpolated
      features, later the output of the previous block) and concatenating those to the
      single prediction representation
    - collecting additional information (iou, normalized distance, l2 distance, scale difference)
      for each pair of predictions in the current neighborhood and concatenating it to the corresponding
      prediction representations.
      This additional information is NOT calculated here but has to be passed to the function by
      'neighborhood_additional_info'

    detailed steps:
    - collect all the prediction representations from 'interpolated' corresponding to boxes
      that are in the neighborhood of the current prediction 'anchor_id'
    - then the prediction representation of the current prediction is tiled and concatenated to
      the previously collected prediction representations
    - the 'neighborhood_additional_info' (consists of iou, normalized distance, l2 distance,
      scale difference in that order) is concatenated

    This function is called for each prediction for each lnms block the output is passed through the block

    S: size of neighborhood
    N: number of predictions
    D: size of a single prediction
    
    Parameters
    ----------
    use_joint_features: bool
        creates a bounding box spanning the neighboring boxes and interpolates the features for that
    use_cross_feature: bool
        creates a cross shape out of two intersecting bounding boxes spanning from the center until specified end or
        image boundary
    anchor_id: int
        id of current prediction in [0, N-1]
    interpolated: tensor (float32)
        N x D interpolated features
    neighborhood_indeces: tensor (int32)
        S x 2 indeces of predictions in neighborhood
    neighborhood_additional_info: tensor (float32)
        S x F additional information of neighborhood
    use_image_features: bool
        boolean whether image features should be added to neighborhood

    Returns
    -------
    concatenated_neighborhood: tensor (float32)
        S x 2D ragged tensor of assembled neighborhood
    """
    anchor_id = tf.cast(anchor_id, tf.int32)

    # collect image features
    if use_image_features:
        # TODO implement joint feature, and cross feature
        neighborhood = tf.reshape(tf.gather(interpolated, neighborhood_indeces),
                                  [tf.size(neighborhood_indeces), tf.shape(interpolated)[1]])
        tiled_pred = tf.tile(tf.expand_dims(interpolated[anchor_id], axis=0), (tf.shape(neighborhood)[0], 1))
    elif False:
        # TODO implement interpolate joint feature representation
        pass
    else:
        neighborhood = tf.reshape(tf.gather(interpolated[:, 0], neighborhood_indeces),
                                  [tf.size(neighborhood_indeces), 1])
        tiled_pred = tf.tile(tf.reshape(interpolated[anchor_id, 0], (1, 1)), (tf.shape(neighborhood)[0], 1))

    concatenated_neighborhood = tf.concat([neighborhood, tiled_pred, neighborhood_additional_info], axis=1)

    return concatenated_neighborhood


def single_neighborhood_additional_info(anchor_id, ious, rpn_boxes_positive,
                                        normalization_factor: float,
                                        threshold: Union[float, tf.Tensor]):
    """
    assembles additional information of the neighborhood of a single prediction 'anchor_id'
    also assembles the neighborhood_indexes

    detailed steps:
    - find the boxes that have an iou of bigger then 'threshold' with the current box 'anchor_id'
    - store the indexes of those boxes in a list 'neighborhood_indexes'
    - collect all the boxes that are in the neighborhood
    - collect the iou of the boxes in the neighborhood with the current box 'anchor_id'
    - calculate the distances (in x and y direction separately) of the boxes and the current box
    - calculate the l2 distance from that
    - calculate the scale difference
    - concatenate all the collected information in 'additional_info'


    S: size of neighborhood
    N: number of predictionsadditional_info
    D: size of a single prediction

    Parameters
    ----------
    anchor_id:
        id of current prediction
    ious:
        N x N ious of predictions with other predictions
    rpn_boxes_positive:
        N x 4 x,y,w,h of the boxes or centroids
    normalization_factor:
        distance normalization devider
    threshold:
        threshold for neighborhood

    Returns
    -------
    additional_info: tensor (float32)
        for each pair of boxes in the neighborhood of the current anchor the following additional information
        is presented in a tensor:
        - iou
        - normalized distance
        - l2 distance
        - scale difference
    neighborhood_indexes: tensor (float32)
        indexes of the other anchors that are in the neighborhood of the current anchor
    """
    anchor_id = tf.cast(anchor_id, tf.int32)

    # find indexes of the boxes that are in the neighborhood of the current box
    neighborhood_indexes = tf.where(tf.greater(ious[:, anchor_id], threshold))

    # collect the boxes that are in the neighborhood of the current prediction
    neighborhood_boxes = tf.gather(rpn_boxes_positive, neighborhood_indexes)
    x, y = tf.shape(neighborhood_boxes)[0], tf.shape(neighborhood_boxes)[2]
    neighborhood_boxes = tf.reshape(neighborhood_boxes, (x, y))

    # calculate additional information for the neighborhood
    collected_ious = tf.gather(ious[:, anchor_id], neighborhood_indexes)
    distances = (neighborhood_boxes[:, :2] - rpn_boxes_positive[anchor_id, :2])
    normalized_distance = tf.abs(distances) / (normalization_factor + 1e-8)
    l2_distance = tf.expand_dims(tf.reduce_sum(distances ** 2, axis=1) ** (1 / 2) / (normalization_factor + 1e-8),
                                 axis=1)

    # concatenate neighborhood vector representations
    scale_difference = kb.log(kb.abs(neighborhood_boxes[:, 2:] / (rpn_boxes_positive[anchor_id, 2:] + 1e-8)) + 1e-8)
    additional_info = tf.concat([collected_ious, normalized_distance,
                                 l2_distance, scale_difference], axis=1)
    return additional_info, neighborhood_indexes


def all_neighborhoods_additional_info(rpn_boxes_positive, prediction_ids, normalization_factor: float,
                                      threshold: Union[float, tf.Tensor], use_distance: bool = False):
    """
    collect additional info and indexes for all neighborhoods by running the method
    'single_neighborhood_additional_info' for each prediction and reformatting the collected data

    neighborhoods of different predictions have different sizes, but RaggedTensors are not ideal.
    Therefore, some reformatting has to done and the neighborhood size has to be stored

    detailed steps:
    - calculate iou for all predictions
    - iterate for each prediction
        + collect additional info and neighborhood indexes with 'single_neighborhood_additional_info'
        + save size of neighborhood (e.g. number of predictions in the neighborhood) in 'neighborhood_sizes'
        + concatenate additional info and indexes for current neighborhood to already collected ones

    there is no clear way to differentiate between neighborhoods in the 'neighborhoods_additional_info' or
    'neighborhoods_indexes' tensors. To find out what belongs to which neighborhood the tensor
    'neighborhood_sizes' has to be used.

    S: size of neighborhood
    N: number of predictions
    D: size of a single prediction

    Parameters
    ----------
    rpn_boxes_positive: tensor (float32)
        N x 4 x,y,w,h of the boxes or centroids
    prediction_ids: tensor (int32)
        list of ids
    normalization_factor: float
        distance normalization divider
    threshold: float
        threshold for neighborhood
    use_distance: bool
        use distance for neighborhood assembly instead of iou

    Returns
    -------
    neighborhood_sizes: tensor (int32)
        the sizes of the neighborhoods of each anchor
    neighborhoods_additional_info: tensor (float32)
        the additional information of each neighborhood collected in one tensor
    neighborhoods_indexes: tensor (int32)
        the indexes of the anchors in each neighborhood
    self_indexes: tensor (int32)
        for each neighborhood the index of the main box is tiled by the number of other predictions in
        the neighborhood and concatenated with the others
    """

    # calculate distance or ious
    if use_distance:
        centroids = rpn_boxes_positive[:, :2] + rpn_boxes_positive[:, 2:]/2
        dist_x = (tf.expand_dims(centroids[:, 0], axis=1) - tf.expand_dims(centroids[:, 0], axis=0))
        dist_y = (tf.expand_dims(centroids[:, 1], axis=1) - tf.expand_dims(centroids[:, 1], axis=0))
        ious = 1/(dist_x**2 + dist_y**2)
    else:
        ious = iou(rpn_boxes_positive, rpn_boxes_positive)

    # assemble initial information
    neighborhood_sizes = tf.TensorArray(tf.int32, size=tf.shape(rpn_boxes_positive)[0])
    neighborhoods_additional_info, neighborhoods_indexes = single_neighborhood_additional_info(
        prediction_ids[0], ious, rpn_boxes_positive, normalization_factor, threshold)

    current_neighborhood_size = tf.shape(neighborhoods_additional_info)[0]
    neighborhood_sizes = neighborhood_sizes.write(0, current_neighborhood_size)
    # self_indexes = tf.expand_dims(tf.tile(0, [current_neighborhood_size]), axis=1)
    self_indexes = tf.expand_dims(tf.tile(tf.constant([0]), [current_neighborhood_size]), axis=1)

    # assemble neighborhoods for each prediction
    for x in prediction_ids[1:]:
        tf.autograph.experimental.set_loop_options(
            shape_invariants=[(neighborhoods_additional_info, tf.TensorShape([None, None])),
                              (neighborhoods_indexes, tf.TensorShape([None, None])),
                              (self_indexes, tf.TensorShape([None, None]))])
        new_neighborhood, new_indexes = single_neighborhood_additional_info(
            x, ious, rpn_boxes_positive, normalization_factor, threshold)

        current_neighborhood_size = tf.shape(new_neighborhood)[0]
        neighborhood_sizes = neighborhood_sizes.write(x, current_neighborhood_size)

        neighborhoods_additional_info = tf.concat([neighborhoods_additional_info, new_neighborhood], axis=0)
        neighborhoods_indexes = tf.concat([neighborhoods_indexes, new_indexes], axis=0)

        current_self_indexes = tf.expand_dims(tf.tile(tf.expand_dims(x, axis=0), [current_neighborhood_size]), axis=1)
        self_indexes = tf.concat([self_indexes, current_self_indexes], axis=0)
    neighborhood_sizes = neighborhood_sizes.stack()

    return neighborhood_sizes, neighborhoods_additional_info, neighborhoods_indexes, self_indexes



import tensorflow as tf


class FalseNegativeRate(tf.keras.metrics.Metric):
    """This metric measures false negative rate at specified thresholds to assess
    performance of the binary classifications.
    
    Attributes
    ----------
    thresholds: float32
        Thresholds in range [0, 1]. Default value 0.5.
    false_negatives: float32
        Running sum of false negatives observed at each threshold.
    positives: float32
        Running sum of true positives.
    """

    def __init__(self, thresholds=None, name='fnr', **kwargs):
        super(FalseNegativeRate, self).__init__(name=name, **kwargs)
        if thresholds is None:
            self.thresholds = tf.convert_to_tensor([0.5])
        else:
            self.thresholds = tf.convert_to_tensor(thresholds)
        self.false_negatives = self.add_weight(name='false_negatives', initializer='zeros',
                                               shape=(tf.size(self.thresholds),))
        self.positives = self.add_weight(name='positives', initializer='zeros')

    def reset_state(self):
        self.false_negatives.assign(tf.zeros(tf.shape(self.false_negatives), tf.float32))
        self.positives.assign(0.)
        
    def update_state(self, y_true, y_pred, sample_weight=None):
        
        #generate binary matrix of false positives
        binary = tf.less_equal(tf.reshape(y_pred, (tf.size(y_pred), 1)),
                               self.thresholds)
        positives = tf.equal(tf.cast(y_true, tf.bool), True)
        false_negatives = tf.logical_and(tf.reshape(positives, (tf.size(y_true), 1)),
                                         binary)
        
        #cast to float for possible weighting
        positives = tf.cast(positives, tf.float32)
        false_negatives = tf.cast(false_negatives, tf.float32)
        
        #weight samples
        if sample_weight is not None:
            sample_weight = tf.cast(sample_weight, tf.float32)
            sample_weight = tf.broadcast_to(sample_weight, (tf.size(y_true), 1))
            negatives = tf.multiply(positives, sample_weight)
            false_negatives = tf.multiply(false_negatives, sample_weight)
                
        #reduce along sample axes
        self.positives.assign_add(tf.reduce_sum(positives))
        self.false_negatives.assign_add(tf.reduce_sum(false_negatives, axis=0))

    def result(self):
        return tf.cond(tf.equal(tf.reduce_sum(self.positives), 0.0),
                       lambda: 0.0,
                       lambda: self.false_negatives / self.positives)
    

class FalsePositiveRate(tf.keras.metrics.Metric):
    """This metric measures false positive rate at specified thresholds to assess
    performance of the binary classifications.
    
    Attributes
    ----------
    thresholds: float32
        Thresholds in range [0, 1]. Default value 0.5.
    false_positives: float32
        Running sum of false positives observed at each threshold.
    negatives: float32
        Running sum of true negatives.
    """
    
    def __init__(self, thresholds=None, name='fpr', **kwargs):
        super(FalsePositiveRate, self).__init__(name=name, **kwargs)
        if thresholds is None:
            self.thresholds = tf.convert_to_tensor([0.5])
        else:
            self.thresholds = tf.convert_to_tensor(thresholds)
        self.false_positives = self.add_weight(name='false_positives', initializer='zeros',
                                               shape=(tf.size(self.thresholds),))
        self.negatives = self.add_weight(name='negatives', initializer='zeros')
        
    def reset_state(self):
        self.false_positives.assign(tf.zeros(tf.shape(self.false_positives), tf.float32))
        self.negatives.assign(0.)

    def update_state(self, y_true, y_pred, sample_weight=None):
        
        #generate binary matrix of false positives
        binary = tf.greater(tf.reshape(y_pred, (tf.size(y_pred), 1)),
                            self.thresholds)
        negatives = tf.equal(tf.cast(y_true, tf.bool), False)
        false_positives = tf.logical_and(tf.reshape(negatives, (tf.size(y_true), 1)),
                                         binary)
        
        #cast to float for possible weighting
        negatives = tf.cast(negatives, tf.float32)
        false_positives = tf.cast(false_positives, tf.float32)
        
        #weight samples
        if sample_weight is not None:
            sample_weight = tf.cast(sample_weight, tf.float32)
            sample_weight = tf.broadcast_to(sample_weight, (tf.size(y_true), 1))
            negatives = tf.multiply(negatives, sample_weight)
            false_positives = tf.multiply(false_positives, sample_weight)
                
        #reduce along sample axes
        self.negatives.assign_add(tf.reduce_sum(negatives))
        self.false_positives.assign_add(tf.reduce_sum(false_positives, axis=0))

    def result(self):
        return tf.cond(tf.equal(tf.reduce_sum(self.negatives), 0.0),
                       lambda: 0.0,
                       lambda: self.false_positives / self.negatives)
